# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import decimal

import octobot_commons.logging as logging

import octobot_trading.errors as errors
import octobot_trading.constants as constants
import octobot_trading.personal_data.portfolios.channel.balance as portfolios_channel
import octobot_trading.exchange_channel as exchange_channel


class BalanceUpdater(portfolios_channel.BalanceProducer):
    """
    The Balance Update fetch the exchange portfolio and send it to the Balance Channel
    """

    """
    The default balance update refresh time in seconds
    """
    BALANCE_REFRESH_TIME = 666

    """
    The updater related channel name
    """
    CHANNEL_NAME = constants.BALANCE_CHANNEL

    async def start(self) -> None:
        """
        Starts the balance updating process
        """
        while not self.should_stop:
            try:
                await self.fetch_and_push()
                await asyncio.sleep(self.BALANCE_REFRESH_TIME)
            except errors.FailedRequest as e:
                self.logger.warning(str(e))
                # avoid spamming on disconnected situation
                await asyncio.sleep(constants.DEFAULT_FAILED_REQUEST_RETRY_TIME)
            except errors.NotSupported:
                self.logger.warning(
                    f"{self.channel.exchange_manager.exchange_name} is not supporting updates"
                )
                await self.pause()
            except Exception as e:
                self.logger.error(f"Failed to update balance : {e}")

    async def fetch_and_push(self):
        await self.push((await self.fetch_portfolio()))

    async def fetch_portfolio(self):
        """
        Fetch portfolio from exchange
        """
        return await self.channel.exchange_manager.exchange.get_balance()

    async def resume(self) -> None:
        """
        Resume updater process
        """
        await super().resume()
        if not self.is_running:
            await self.run()


class BalanceProfitabilityUpdater(portfolios_channel.BalanceProfitabilityProducer):
    """
    The Balance Profitability Updater triggers the portfolio profitability calculation
    by subscribing to Mark price and Balance channel updates
    """

    """
    The updater related channel name
    """
    CHANNEL_NAME = constants.BALANCE_PROFITABILITY_CHANNEL

    def __init__(self, channel):
        super().__init__(channel)
        self.logger = logging.get_logger(self.__class__.__name__)
        self.exchange_personal_data = (
            self.channel.exchange_manager.exchange_personal_data
        )
        self.balance_consumer = None
        self.mark_price_consumer = None

    async def start(self) -> None:
        """
        Starts the balance profitability subscribing process
        """
        self.balance_consumer = await exchange_channel.get_chan(
            constants.BALANCE_CHANNEL, self.channel.exchange_manager.id
        ).new_consumer(self.handle_balance_update)
        self.mark_price_consumer = await exchange_channel.get_chan(
            constants.MARK_PRICE_CHANNEL, self.channel.exchange_manager.id
        ).new_consumer(self.handle_mark_price_update)

    async def stop(self) -> None:
        """
        Stop and remove the balance profitability consumers
        """
        await super().stop()
        try:
            await exchange_channel.get_chan(
                constants.BALANCE_CHANNEL, self.channel.exchange_manager.id
            ).remove_consumer(self.balance_consumer)
        except KeyError:
            # balance channel might already be stopped and removed from available channels
            pass
        try:
            await exchange_channel.get_chan(
                constants.MARK_PRICE_CHANNEL, self.channel.exchange_manager.id
            ).remove_consumer(self.mark_price_consumer)
        except KeyError:
            # balance channel might already be stopped and removed from available channels
            pass
        self.balance_consumer = None
        self.mark_price_consumer = None

    async def handle_balance_update(
        self, exchange: str, exchange_id: str, balance: dict
    ) -> None:
        """
        Balance channel consumer callback
        :param exchange: the exchange name
        :param exchange_id: the exchange id
        :param balance: the balance dict
        """
        try:
            await self.exchange_personal_data.handle_portfolio_profitability_update(
                balance=balance, mark_price=None, symbol=None
            )
        except Exception as e:
            self.logger.exception(e, True, f"Fail to handle balance update : {e}")

    async def handle_mark_price_update(
        self,
        exchange: str,
        exchange_id: str,
        cryptocurrency: str,
        symbol: str,
        mark_price: decimal.Decimal,
    ) -> None:
        """
        Mark price channel consumer callback
        :param exchange: the exchange name
        :param exchange_id: the exchange id
        :param cryptocurrency: the related currency
        :param symbol: the related symbol
        :param mark_price: the mark price
        """
        try:
            await self.exchange_personal_data.handle_portfolio_profitability_update(
                symbol=symbol, mark_price=mark_price, balance=None
            )
        except Exception as e:
            self.logger.exception(e, True, f"Fail to handle mark price update : {e}")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import os
import pytest

import octobot_trading.constants as constants
from tests.test_utils.random_numbers import decimal_random_quantity, decimal_random_price

from tests.exchanges import backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_get_current_crypto_currencies_values(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    assert portfolio_value_holder.get_current_crypto_currencies_values() == \
           {'BTC': constants.ONE, 'USDT': constants.ZERO}
    portfolio_manager.portfolio.update_portfolio_from_balance({
        'BTC': {'available': decimal_random_quantity(), 'total': decimal_random_quantity()},
        'ETH': {'available': decimal_random_quantity(), 'total': decimal_random_quantity()},
        'XRP': {'available': decimal_random_quantity(), 'total': decimal_random_quantity()},
        'DOT': {'available': decimal_random_quantity(), 'total': decimal_random_quantity()},
        'MATIC': {'available': decimal_random_quantity(), 'total': decimal_random_quantity()},
        'USDT': {'available': decimal_random_quantity(), 'total': decimal_random_quantity()}
    }, True)
    portfolio_manager.handle_balance_updated()

    assert portfolio_value_holder.get_current_crypto_currencies_values() == {
        'BTC': constants.ONE,
        'ETH': constants.ZERO,
        'XRP': constants.ZERO,
        'DOT': constants.ZERO,
        'MATIC': constants.ZERO,
        'USDT': constants.ZERO
    }

    exchange_manager.client_symbols.append("MATIC/BTC")
    exchange_manager.client_symbols.append("XRP/BTC")
    if not os.getenv('CYTHON_IGNORE'):
        portfolio_value_holder.value_converter.missing_currency_data_in_exchange.remove("XRP")
        portfolio_manager.handle_mark_price_update("XRP/BTC", decimal.Decimal("0.005"))
        exchange_manager.client_symbols.append("DOT/BTC")
        portfolio_value_holder.value_converter.missing_currency_data_in_exchange.remove("DOT")
        portfolio_manager.handle_mark_price_update("DOT/BTC", decimal.Decimal("0.05"))
        exchange_manager.client_symbols.append("BTC/USDT")

        assert portfolio_value_holder.get_current_crypto_currencies_values() == {
            'BTC': constants.ONE,
            'ETH': constants.ZERO,
            'XRP': decimal.Decimal("0.005"),
            'DOT': decimal.Decimal("0.05"),
            'MATIC': constants.ZERO,
            'USDT': constants.ZERO
        }
        matic_btc_price = decimal_random_price(max_value=decimal.Decimal(0.05))
        portfolio_value_holder.value_converter.missing_currency_data_in_exchange.remove("MATIC")
        portfolio_manager.handle_mark_price_update("MATIC/BTC", matic_btc_price)
        assert portfolio_value_holder.get_current_crypto_currencies_values() == {
            'BTC': constants.ONE,
            'ETH': constants.ZERO,
            'XRP': decimal.Decimal("0.005"),
            'DOT': decimal.Decimal("0.05"),
            'MATIC': matic_btc_price,
            'USDT': constants.ZERO
        }
        usdt_btc_price = decimal_random_price(max_value=decimal.Decimal('0.01'))
        portfolio_value_holder.value_converter.missing_currency_data_in_exchange.remove("USDT")
        portfolio_manager.handle_mark_price_update("BTC/USDT", usdt_btc_price)
        assert portfolio_value_holder.get_current_crypto_currencies_values() == {
            'BTC': constants.ONE,
            'ETH': constants.ZERO,
            'XRP': decimal.Decimal("0.005"),
            'DOT': decimal.Decimal("0.05"),
            'MATIC': matic_btc_price,
            'USDT': constants.ONE / usdt_btc_price
        }
        eth_btc_price = decimal_random_price(max_value=constants.ONE)
        exchange_manager.client_symbols.append("ETH/BTC")
        portfolio_value_holder.value_converter.missing_currency_data_in_exchange.remove("ETH")
        portfolio_manager.handle_mark_price_update("ETH/BTC", eth_btc_price)
        assert portfolio_value_holder.get_current_crypto_currencies_values() == {
            'BTC': constants.ONE,
            'ETH': decimal.Decimal(str(eth_btc_price)),
            'XRP': decimal.Decimal("0.005"),
            'DOT': decimal.Decimal("0.05"),
            'MATIC': matic_btc_price,
            'USDT': constants.ONE / usdt_btc_price
        }


async def test_get_current_holdings_values(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    exchange_manager.client_symbols.append("ETH/BTC")
    portfolio_manager.portfolio.update_portfolio_from_balance({
        'BTC': {'available': decimal.Decimal("10"), 'total': decimal.Decimal("10")},
        'ETH': {'available': decimal.Decimal("100"), 'total': decimal.Decimal("100")},
        'XRP': {'available': decimal.Decimal("10000"), 'total': decimal.Decimal("10000")},
        'USDT': {'available': decimal.Decimal("1000"), 'total': decimal.Decimal("1000")}
    }, True)
    portfolio_manager.handle_balance_updated()
    assert portfolio_value_holder.get_current_holdings_values() == {
        'BTC': decimal.Decimal("10"),
        'ETH': constants.ZERO,
        'XRP': constants.ZERO,
        'USDT': constants.ZERO
    }
    portfolio_manager.handle_mark_price_update("ETH/BTC", 50)
    assert portfolio_value_holder.get_current_holdings_values() == {
        'BTC': decimal.Decimal("10"),
        'ETH': decimal.Decimal("5000"),
        'XRP': constants.ZERO,
        'USDT': constants.ZERO
    }
    portfolio_manager.handle_mark_price_update("XRP/USDT", 2.5)
    assert portfolio_value_holder.get_current_holdings_values() == {
        'BTC': decimal.Decimal("10"),
        'ETH': decimal.Decimal("5000"),
        'XRP': constants.ZERO,
        'USDT': constants.ZERO
    }
    portfolio_manager.handle_mark_price_update("XRP/BTC", decimal.Decimal('0.00001'))
    assert portfolio_value_holder.get_current_holdings_values() == {
        'BTC': decimal.Decimal("10"),
        'ETH': decimal.Decimal("5000"),
        'XRP': constants.ZERO,
        'USDT': constants.ZERO
    }
    if not os.getenv('CYTHON_IGNORE'):
        exchange_manager.client_symbols.append("XRP/BTC")
        portfolio_value_holder.value_converter.missing_currency_data_in_exchange.remove("XRP")
        portfolio_manager.handle_mark_price_update("XRP/BTC", decimal.Decimal('0.00001'))
        assert portfolio_value_holder.get_current_holdings_values() == {
            'BTC': decimal.Decimal(10),
            'ETH': decimal.Decimal(5000),
            'XRP': decimal.Decimal(str(0.1)),
            'USDT': constants.ZERO
        }
        exchange_manager.client_symbols.append("BTC/USDT")
        portfolio_value_holder.value_converter.missing_currency_data_in_exchange.remove("USDT")
        portfolio_manager.handle_mark_price_update("BTC/USDT", 5000)
        assert portfolio_value_holder.get_current_holdings_values() == {
            'BTC': decimal.Decimal(10),
            'ETH': decimal.Decimal(5000),
            'XRP': decimal.Decimal(str(0.1)),
            'USDT': decimal.Decimal(str(0.2))
        }


async def test_get_origin_portfolio_current_value(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    portfolio_manager.handle_profitability_recalculation(True)
    assert portfolio_value_holder.get_origin_portfolio_current_value() == decimal.Decimal(str(10))


async def test_get_origin_portfolio_current_value_with_different_reference_market(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    portfolio_manager.reference_market = "USDT"
    portfolio_manager.handle_profitability_recalculation(True)
    assert portfolio_value_holder.get_origin_portfolio_current_value() == decimal.Decimal(str(1000))


async def test_update_origin_crypto_currencies_values(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    assert portfolio_value_holder.update_origin_crypto_currencies_values("ETH/BTC", decimal.Decimal(str(0.1))) is True
    assert portfolio_value_holder.origin_crypto_currencies_values["ETH"] == decimal.Decimal(str(0.1))
    assert portfolio_value_holder.value_converter.last_prices_by_trading_pair["ETH/BTC"] == decimal.Decimal(str(0.1))
    # ETH is now priced and BTC is the reference market
    assert portfolio_value_holder.update_origin_crypto_currencies_values("ETH/BTC", decimal.Decimal(str(0.1))) is False

    assert portfolio_value_holder.update_origin_crypto_currencies_values("BTC/USDT", decimal.Decimal(str(100))) is True
    assert portfolio_value_holder.origin_crypto_currencies_values["USDT"] == \
           decimal.Decimal(constants.ONE / decimal.Decimal(100))
    assert portfolio_value_holder.value_converter.last_prices_by_trading_pair["BTC/USDT"] == decimal.Decimal(str(100))
    # USDT is now priced and BTC is the reference market
    assert portfolio_value_holder.update_origin_crypto_currencies_values("BTC/USDT", decimal.Decimal(str(100))) is False

    # with bridge pair (DOT/ETH -> ETH/BTC to compute DOT/BTC)
    assert portfolio_value_holder.update_origin_crypto_currencies_values("DOT/ETH", decimal.Decimal(str(0.015))) is True
    assert portfolio_value_holder.origin_crypto_currencies_values["DOT"] == \
           decimal.Decimal(str(0.015)) * decimal.Decimal(str(0.1))
    assert portfolio_value_holder.value_converter.last_prices_by_trading_pair["DOT/ETH"] == decimal.Decimal(str(0.015))
    # USDT is now priced and BTC is the reference market
    assert portfolio_value_holder.update_origin_crypto_currencies_values("DOT/ETH", decimal.Decimal(str(0.015))) \
           is False



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging

import octobot_trading.enums as enums
import octobot_trading.constants as constants


class MarginContract:
    def __init__(self, pair, margin_type,
                 contract_size=constants.ONE,
                 maximum_leverage=constants.ONE,
                 current_leverage=constants.ONE):
        self.pair = pair
        self.margin_type = margin_type

        self.contract_size = contract_size
        self.maximum_leverage = maximum_leverage
        self.current_leverage = current_leverage

        self.risk_limit = {}

    def is_isolated(self):
        """
        Margin in isolated margin mode is independent for each trading pair
        Margin in cross margin mode is shared among the user’s margin account
        :return: True if the contract use isolation margin
        """
        return self.margin_type is enums.MarginType.ISOLATED

    def check_leverage_update(self, new_leverage):
        """
        Check if the new leverage value is matching requirements (0 < maximum_leverage)
        :param new_leverage: the leverage value to check
        :return: True if valid
        """
        return 0 < new_leverage if self.maximum_leverage is None else 0 < new_leverage <= self.maximum_leverage

    def set_current_leverage(self, new_leverage):
        """
        Set the contract current leverage value
        :param new_leverage: the new leverage value
        """
        if self.check_leverage_update(new_leverage):
            self.current_leverage = new_leverage

    def set_margin_type(self, is_isolated=True, is_cross=False):
        """
        Set the contract margin type
        :param is_isolated: should be True if the margin type is isolated
        :param is_cross: should be True if the margin type is cross
        """
        self.margin_type = enums.MarginType.ISOLATED if is_isolated and not is_cross else enums.MarginType.CROSS

    def is_handled_contract(self):
        """
        :return: False if this contract can't be traded in OctoBot
        """
        return True

    def update_from_position(self, raw_position) -> bool:
        changed = False
        margin_type = raw_position.get(enums.ExchangeConstantsPositionColumns.MARGIN_TYPE.value, None)
        if margin_type is not None and self.margin_type != margin_type:
            self.set_margin_type(
                is_isolated=margin_type is enums.MarginType.ISOLATED,
                is_cross=margin_type is enums.MarginType.CROSS
            )
            logging.get_logger(str(self)).debug(f"Changed margin type to {margin_type}")
            changed = True
        return changed



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.constants as constants
import octobot_trading.personal_data.portfolios.portfolio as portfolio_class


class SubPortfolio(portfolio_class.Portfolio):
    DEFAULT_SUB_PORTFOLIO_PERCENT = decimal.Decimal("0.5")

    def __init__(self, config, trader, parent_portfolio, percent, is_relative=True):
        self.parent_portfolio = parent_portfolio
        self.is_relative = is_relative
        self.set_percent(percent)
        super().__init__(config, trader)

    # overwrite parent update_portfolio_balance
    def update_portfolio_from_balance(self, balance, force_replace=True):
        modified = self.parent_portfolio.update_portfolio_from_balance(balance, force_replace=force_replace)
        self.update_from_parent()
        return modified

    def update_from_parent(self): # TODO
        # # get the current portfolio if percent is relative or if we can't use the origin portfolio
        # if self.is_relative or not self.trader.get_trades_manager().get_origin_portfolio():
        #     balance = self.parent_portfolio.get_portfolio()
        #
        # # the percent is applied to the origin portfolio (when not relative)
        # else:
        #     balance = self.trader.get_trades_manager().get_origin_portfolio()
        #
        # # calculate for each currency the new quantity
        # self.portfolio = {currency: {PORTFOLIO_AVAILABLE: balance[currency][PORTFOLIO_AVAILABLE] * self.percent,
        #                              PORTFOLIO_TOTAL: balance[currency][PORTFOLIO_TOTAL] * self.percent}
        #                   for currency in balance}
        pass

    def set_percent(self, percent):
        if percent and percent > constants.ZERO:
            self.percent = percent
        else:
            self.percent = self.DEFAULT_SUB_PORTFOLIO_PERCENT

    def update_portfolio_from_filled_order(self, order):
        super().update_portfolio_from_filled_order(order)
        self.parent_portfolio.update_portfolio_from_filled_order(order)

    def update_portfolio_available(self, order, is_new_order=False):
        super().update_portfolio_available(order, is_new_order=is_new_order)
        self.parent_portfolio.update_portfolio_available(order, is_new_order=is_new_order)

    def reset_portfolio_available(self, reset_currency=None, reset_quantity=None):
        super().reset_portfolio_available(reset_currency=reset_currency, reset_quantity=reset_quantity)
        self.parent_portfolio.reset_portfolio_available(reset_currency=reset_currency, reset_quantity=reset_quantity)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
#  License along with this library.
import decimal

import octobot_trading.constants as constants
import octobot_trading.personal_data.portfolios.assets.margin_asset as margin_asset

ASSET_CURRENCY_NAME = "BTC"


def test___eq__():
    asset = margin_asset.MarginAsset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ZERO)
    assert asset == margin_asset.MarginAsset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ZERO)
    assert not asset == margin_asset.MarginAsset(ASSET_CURRENCY_NAME, constants.ONE_HUNDRED, constants.ZERO)
    assert not asset == margin_asset.MarginAsset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ONE_HUNDRED)
    assert not asset == margin_asset.MarginAsset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ZERO,
                                                 borrowed=constants.ONE_HUNDRED)
    assert asset == margin_asset.MarginAsset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ZERO,
                                             borrowed=constants.ZERO)
    assert not asset == margin_asset.MarginAsset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ZERO,
                                                 interest=constants.ONE_HUNDRED)
    assert not asset == margin_asset.MarginAsset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ZERO,
                                                 locked=constants.ONE_HUNDRED)


def test_update():
    asset = margin_asset.MarginAsset(ASSET_CURRENCY_NAME,
                                     available=constants.ZERO, total=constants.ZERO,
                                     borrowed=constants.ZERO, interest=constants.ZERO, locked=constants.ZERO)
    assert not asset.update(available=constants.ZERO, total=constants.ZERO)
    assert asset.update(available=constants.ZERO, total=decimal.Decimal(5))
    assert asset.total == decimal.Decimal(5)
    assert asset.update(available=constants.ZERO, borrowed=decimal.Decimal(15))
    assert asset.borrowed == decimal.Decimal(15)
    assert asset.total == decimal.Decimal(5)
    assert asset.available == constants.ZERO


def test_set():
    asset = margin_asset.MarginAsset(ASSET_CURRENCY_NAME,
                                     available=constants.ZERO, total=constants.ZERO,
                                     borrowed=constants.ZERO, interest=constants.ZERO, locked=constants.ZERO)
    assert not asset.set(available=constants.ZERO, total=constants.ZERO)
    assert asset.set(available=decimal.Decimal(5), total=decimal.Decimal(5), borrowed=decimal.Decimal(2),
                     interest=decimal.Decimal(2), locked=decimal.Decimal(7))
    assert not asset.set(available=decimal.Decimal(5), total=decimal.Decimal(5), borrowed=decimal.Decimal(2),
                         interest=decimal.Decimal(2), locked=decimal.Decimal(7))
    assert asset.available == decimal.Decimal(5)
    assert asset.total == decimal.Decimal(5)
    assert asset.borrowed == decimal.Decimal(2)
    assert asset.locked == decimal.Decimal(7)
    assert asset.interest == decimal.Decimal(2)


def test_restore_available():
    asset = margin_asset.MarginAsset(ASSET_CURRENCY_NAME, available=constants.ONE, total=constants.ONE_HUNDRED)
    assert not asset.available == asset.total
    asset.restore_available()
    assert asset.available == asset.total == constants.ONE_HUNDRED


def test_reset():
    asset = margin_asset.MarginAsset(ASSET_CURRENCY_NAME, available=constants.ONE_HUNDRED,
                                     total=constants.ONE_HUNDRED, borrowed=constants.ONE_HUNDRED,
                                     interest=constants.ONE_HUNDRED, locked=constants.ONE_HUNDRED)
    assert asset.available == asset.total == asset.interest == asset.borrowed == asset.locked == constants.ONE_HUNDRED
    asset.reset()
    assert asset.available == asset.total == asset.interest == asset.borrowed == asset.locked == constants.ZERO



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from tests import event_loop
from octobot_trading.api.channels import subscribe_to_ohlcv_channel, subscribe_to_trades_channel, \
    subscribe_to_order_channel

from tests.exchanges import exchange_manager

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def ohlcv_callback(
        exchange: str,
        exchange_id: str,
        cryptocurrency: str,
        symbol: str,
        time_frame,
        candle
):
    pass


async def trades_callback(exchange: str, exchange_id: str, cryptocurrency: str, symbol: str, trade, old_trade):
    pass


async def order_callback(self, exchange, exchange_id, cryptocurrency, symbol, order, update_type, is_from_bot):
    pass


async def test_subscribe_to_ohlcv_channel(exchange_manager):
    await subscribe_to_ohlcv_channel(ohlcv_callback, exchange_manager.id)


async def test_subscribe_to_trades_channel(exchange_manager):
    await subscribe_to_trades_channel(trades_callback, exchange_manager.id)


async def test_subscribe_to_order_channel(exchange_manager):
    await subscribe_to_order_channel(order_callback, exchange_manager.id)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import numpy as np

import octobot_commons.data_util as data_util
import octobot_commons.enums as enums


def get_symbol_close_candles(symbol_data, time_frame, limit, include_in_construction):
    tf = enums.TimeFrames(time_frame)
    if include_in_construction:
        return _add_in_construction_data(
            symbol_data.symbol_candles[tf].get_symbol_close_candles(limit),
            symbol_data,
            tf,
            enums.PriceIndexes.IND_PRICE_CLOSE.value)
    return symbol_data.symbol_candles[tf].get_symbol_close_candles(limit)


def get_symbol_open_candles(symbol_data, time_frame, limit, include_in_construction):
    tf = enums.TimeFrames(time_frame)
    if include_in_construction:
        return _add_in_construction_data(
            symbol_data.symbol_candles[tf].get_symbol_open_candles(limit),
            symbol_data,
            tf,
            enums.PriceIndexes.IND_PRICE_OPEN.value)
    return symbol_data.symbol_candles[tf].get_symbol_open_candles(limit)


def get_symbol_high_candles(symbol_data, time_frame, limit, include_in_construction):
    tf = enums.TimeFrames(time_frame)
    if include_in_construction:
        return _add_in_construction_data(
            symbol_data.symbol_candles[tf].get_symbol_high_candles(limit),
            symbol_data,
            tf,
            enums.PriceIndexes.IND_PRICE_HIGH.value)
    return symbol_data.symbol_candles[tf].get_symbol_high_candles(limit)


def get_symbol_low_candles(symbol_data, time_frame, limit, include_in_construction):
    tf = enums.TimeFrames(time_frame)
    if include_in_construction:
        return _add_in_construction_data(
            symbol_data.symbol_candles[tf].get_symbol_low_candles(limit),
            symbol_data,
            tf,
            enums.PriceIndexes.IND_PRICE_LOW.value)
    return symbol_data.symbol_candles[tf].get_symbol_low_candles(limit)


def get_symbol_volume_candles(symbol_data, time_frame, limit, include_in_construction):
    tf = enums.TimeFrames(time_frame)
    if include_in_construction:
        return _add_in_construction_data(
            symbol_data.symbol_candles[tf].get_symbol_volume_candles(limit),
            symbol_data,
            tf,
            enums.PriceIndexes.IND_PRICE_VOL.value)
    return symbol_data.symbol_candles[tf].get_symbol_volume_candles(limit)


def get_symbol_time_candles(symbol_data, time_frame, limit, include_in_construction):
    tf = enums.TimeFrames(time_frame)
    if include_in_construction:
        return _add_in_construction_data(
            symbol_data.symbol_candles[tf].get_symbol_time_candles(limit),
            symbol_data,
            tf,
            enums.PriceIndexes.IND_PRICE_TIME.value)
    return symbol_data.symbol_candles[tf].get_symbol_time_candles(limit)


def get_candle_as_list(candle_arrays_dict: dict, candle_index: int) -> list:
    candle = [None] * len(enums.PriceIndexes)
    candle[enums.PriceIndexes.IND_PRICE_TIME.value] = candle_arrays_dict[enums.PriceIndexes.IND_PRICE_TIME.value][candle_index]
    candle[enums.PriceIndexes.IND_PRICE_OPEN.value] = candle_arrays_dict[enums.PriceIndexes.IND_PRICE_OPEN.value][candle_index]
    candle[enums.PriceIndexes.IND_PRICE_HIGH.value] = candle_arrays_dict[enums.PriceIndexes.IND_PRICE_HIGH.value][candle_index]
    candle[enums.PriceIndexes.IND_PRICE_LOW.value] = candle_arrays_dict[enums.PriceIndexes.IND_PRICE_LOW.value][candle_index]
    candle[enums.PriceIndexes.IND_PRICE_CLOSE.value] = candle_arrays_dict[enums.PriceIndexes.IND_PRICE_CLOSE.value][candle_index]
    candle[enums.PriceIndexes.IND_PRICE_VOL.value] = candle_arrays_dict[enums.PriceIndexes.IND_PRICE_VOL.value][candle_index]
    return candle


def _add_in_construction_data(candles, symbol_data, time_frame, data_type):
    try:
        return np.array(data_util.shift_value_array(candles,
                                                    fill_value=symbol_data.symbol_klines[time_frame].kline[data_type]),
                        dtype=np.float64)
    except KeyError:
        return candles



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import pytest
import octobot_trading.constants as constants
import octobot_commons.constants as commons_constants
from octobot_trading.enums import FeePropertyColumns, ExchangeConstantsMarketPropertyColumns, TraderOrderType
from octobot_trading.api.exchange import cancel_ccxt_throttle_task

# Import required fixtures
from tests import event_loop
from tests.exchanges import backtesting_trader, backtesting_config, backtesting_config, backtesting_exchange_manager, \
    DEFAULT_BACKTESTING_SYMBOL, DEFAULT_BACKTESTING_TF, DEFAULT_BACKTESTING_SPLIT_SYMBOL, DEFAULT_BACKTESTING_CURRENCY, \
    DEFAULT_BACKTESTING_MARKET, fake_backtesting

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


def _get_start_index_for_timeframe(nb_candles, min_limit, timeframe_multiplier):
    return int(nb_candles - (nb_candles - min_limit) / timeframe_multiplier) - 1


def _assert_fee(fee, currency, price, rate, fee_type):
    assert fee[FeePropertyColumns.CURRENCY.value] == currency
    assert fee[FeePropertyColumns.COST.value] == price
    assert fee[FeePropertyColumns.RATE.value] == rate
    assert fee[FeePropertyColumns.TYPE.value] == fee_type
    assert fee[FeePropertyColumns.IS_FROM_EXCHANGE.value] == False


async def test_is_authenticated(backtesting_trader):
    _, exchange_manager, trader_inst = backtesting_trader
    assert not exchange_manager.exchange.authenticated()


async def test_get_uniform_timestamp(backtesting_trader):
    _, exchange_manager, trader_inst = backtesting_trader
    assert exchange_manager.exchange.get_uniform_timestamp(1e8) == 1e5


async def test_get_max_handled_pair_with_time_frame(backtesting_trader):
    _, exchange_manager, _ = backtesting_trader
    assert exchange_manager.exchange.get_max_handled_pair_with_time_frame() == \
           constants.INFINITE_MAX_HANDLED_PAIRS_WITH_TIMEFRAME


async def test_get_split_pair_from_exchange(backtesting_trader):
    _, exchange_manager, trader_inst = backtesting_trader
    assert exchange_manager.exchange.get_split_pair_from_exchange(DEFAULT_BACKTESTING_SYMBOL) == \
           DEFAULT_BACKTESTING_SPLIT_SYMBOL


async def test_get_pair_cryptocurrency(backtesting_trader):
    _, exchange_manager, trader_inst = backtesting_trader
    return exchange_manager.exchange.get_pair_cryptocurrency(DEFAULT_BACKTESTING_SYMBOL) == DEFAULT_BACKTESTING_CURRENCY


async def test_get_trade_fee(backtesting_trader):
    _, exchange_manager, trader_inst = backtesting_trader

    # force fees
    exchange_manager.config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_SIMULATOR_FEES] = {
        commons_constants.CONFIG_SIMULATOR_FEES_MAKER: 0.05,
        commons_constants.CONFIG_SIMULATOR_FEES_TAKER: 0.1
    }

    buy_market_fee = exchange_manager.exchange.get_trade_fee(DEFAULT_BACKTESTING_SYMBOL, TraderOrderType.BUY_MARKET, 10,
                                                             100, ExchangeConstantsMarketPropertyColumns.TAKER.value)
    _assert_fee(buy_market_fee, DEFAULT_BACKTESTING_CURRENCY, decimal.Decimal("0.01"), 0.001,
                ExchangeConstantsMarketPropertyColumns.TAKER.value)

    sell_market_fee = exchange_manager.exchange.get_trade_fee(
        DEFAULT_BACKTESTING_SYMBOL, TraderOrderType.SELL_MARKET, 10, 100,
        ExchangeConstantsMarketPropertyColumns.TAKER.value)
    _assert_fee(sell_market_fee, DEFAULT_BACKTESTING_MARKET, constants.ONE, 0.001,
                ExchangeConstantsMarketPropertyColumns.TAKER.value)


async def test_stop(backtesting_trader):
    _, exchange_manager, trader_inst = backtesting_trader
    await exchange_manager.exchange.stop()
    cancel_ccxt_throttle_task()
    assert not exchange_manager.exchange.backtesting
    assert not exchange_manager.exchange.exchange_importers


#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.constants as trading_constants
import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.supervisors.abstract_supervisor as abstract_supervisor


class AbstractPortfolioSupervisor(abstract_supervisor.AbstractSupervisor):
    @staticmethod
    def is_backtestable() -> bool:
        return True

    async def initialize(self) -> None:
        """
        Initialize balance channel consumer
        """
        self.consumers.append(await exchanges_channel.get_chan(
            trading_constants.BALANCE_CHANNEL, self.exchange_manager.id).new_consumer(self.on_balance_update))
        self.consumers.append(await exchanges_channel.get_chan(
            trading_constants.BALANCE_PROFITABILITY_CHANNEL, self.exchange_manager.id).new_consumer(
            self.on_balance_profitability_update))

    async def on_balance_update(self, exchange: str, exchange_id: str, balance):
        """
        Should be overwritten
        """

    async def on_balance_profitability_update(
            self,
            exchange: str,
            exchange_id: str,
            profitability,
            profitability_percent,
            market_profitability_percent,
            initial_portfolio_current_profitability,
    ):
        """
        Should be overwritten
        """



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import typing
import decimal
import time

import octobot_commons.constants
import octobot_commons.enums as common_enums
import octobot_commons.logging as logging
import octobot_commons.timestamp_util as timestamp_util
import octobot_commons.tentacles_management as tentacles_management

import octobot_trading.constants
import octobot_trading.enums as enums
import octobot_trading.errors as errors


class AbstractExchange(tentacles_management.AbstractTentacle):
    USER_INPUT_TENTACLE_TYPE = common_enums.UserInputTentacleTypes.EXCHANGE
    BUY_STR = enums.TradeOrderSide.BUY.value
    SELL_STR = enums.TradeOrderSide.SELL.value

    # should be overridden locally to match exchange support
    SUPPORTED_ELEMENTS = {
        enums.ExchangeTypes.FUTURE.value: {
            # order that should be self-managed by OctoBot
            enums.ExchangeSupportedElements.UNSUPPORTED_ORDERS.value: [
                enums.TraderOrderType.STOP_LOSS,
                enums.TraderOrderType.STOP_LOSS_LIMIT,
                enums.TraderOrderType.TAKE_PROFIT,
                enums.TraderOrderType.TAKE_PROFIT_LIMIT,
                enums.TraderOrderType.TRAILING_STOP,
                enums.TraderOrderType.TRAILING_STOP_LIMIT
            ],
            # order that can be bundled together to create them all in one request
            # format: dict of bundled types by base order type
            # ex: SUPPORTED_BUNDLED_ORDERS[enums.TraderOrderType.BUY_MARKET] = \
            # [enums.TraderOrderType.STOP_LOSS, enums.TraderOrderType.TAKE_PROFIT]
            enums.ExchangeSupportedElements.SUPPORTED_BUNDLED_ORDERS.value: {},
        },
        enums.ExchangeTypes.SPOT.value: {
            # order that should be self-managed by OctoBot
            enums.ExchangeSupportedElements.UNSUPPORTED_ORDERS.value: [
                enums.TraderOrderType.STOP_LOSS,
                enums.TraderOrderType.STOP_LOSS_LIMIT,
                enums.TraderOrderType.TAKE_PROFIT,
                enums.TraderOrderType.TAKE_PROFIT_LIMIT,
                enums.TraderOrderType.TRAILING_STOP,
                enums.TraderOrderType.TRAILING_STOP_LIMIT
            ],
            # order that can be bundled together to create them all in one request
            # format: dict of bundled types by base order type
            # ex: SUPPORTED_BUNDLED_ORDERS[enums.TraderOrderType.BUY_MARKET] = \
            # [enums.TraderOrderType.STOP_LOSS, enums.TraderOrderType.TAKE_PROFIT]
            enums.ExchangeSupportedElements.SUPPORTED_BUNDLED_ORDERS.value: {},
        }
    }
    ACCOUNTS = {}

    def __init__(self, config, exchange_manager):
        super().__init__()
        self.config = config
        self.exchange_manager = exchange_manager
        self.connector = None
        self.is_initialized = False

        self.tentacle_config = {}

        # Initialized when initializing exchange connector
        self.symbols = set()
        self.time_frames = set()

        # exchange name related attributes
        self.name = self.exchange_manager.exchange_class_string
        self.logger = logging.get_logger(f"{self.__class__.__name__}[{self.name}]")

        # exchange related constants
        self.allowed_time_lag = octobot_trading.constants.DEFAULT_EXCHANGE_TIME_LAG
        self.current_account = enums.AccountTypes.CASH

        self.is_unreachable = False

        if self.exchange_manager.tentacles_setup_config is not None:
            self.load_user_inputs_from_class(self.exchange_manager.tentacles_setup_config, self.tentacle_config)

    async def initialize(self, force=False, **kwargs):
        if not self.is_initialized or force:
            await self.initialize_impl(**kwargs)
            self.is_initialized = True
            return True
        return False

    async def initialize_impl(self):
        """
        Contains the exchange initialization code
        """
        raise NotImplementedError("initialize_impl not implemented")

    async def stop(self) -> None:
        """
        Implement the exchange stopping process
        """
        raise NotImplementedError("stop not implemented")

    @classmethod
    def get_name(cls) -> str:
        """
        :return: the exchange name
        """
        raise NotImplementedError("get_name is not implemented")

    @classmethod
    def is_simulated_exchange(cls) -> bool:
        """
        :return: True if this implementation corresponds to a simulated exchange
        """
        return False

    def authenticated(self) -> bool:
        return False if self.connector is None else self.connector.is_authenticated

    @classmethod
    def is_default_exchange(cls) -> bool:
        """
        :return: True if this implementation corresponds to a default exchange implementation
        """
        return False

    @classmethod
    def is_supporting_exchange(cls, exchange_candidate_name) -> bool:
        """
        :param exchange_candidate_name: the exchange name
        :return: True if this implementation supports the exchange name
        """
        raise NotImplementedError("is_supporting_exchange is not implemented")

    @classmethod
    def is_supporting_sandbox(cls) -> bool:
        """
        :return: True if the exchange is supporting sandbox mode
        """
        return True

    def get_exchange_current_time(self):
        """
        :return: the exchange current time in seconds
        """
        raise NotImplementedError("get_exchange_current_time is not implemented")

    def get_uniform_timestamp(self, timestamp):
        """
        WARNING: does not check if the timestamp is already in the right form, better using get_uniformized_timestamp
        to be sure to keep the right format and uniformize if necessary only
        :param timestamp: the timestamp to uniformize
        :return: the uniformized timestamp
        """
        raise NotImplementedError("get_uniform_timestamp not implemented")

    def get_supported_elements(self, element_key: enums.ExchangeSupportedElements):
        """
        :return: the supported elements such as order type and bundle orders for the current exchange and trading type
        """
        exchange_type_key = enums.ExchangeTypes.FUTURE if self.exchange_manager.is_future \
            else enums.ExchangeTypes.SPOT
        return self.__class__.SUPPORTED_ELEMENTS[exchange_type_key.value][element_key.value]

    def get_market_status(self, symbol, price_example=None, with_fixer=True):
        """
        Return the market status
        :param symbol: the symbol
        :param price_example: a price example to be used in MarketStatusFixer
        :param with_fixer: when True, return a new instance of MarketStatusFixer
        :return: market status dict
        """
        raise NotImplementedError("get_market_status is not implemented")

    async def get_balance(self, **kwargs: dict):
        """
        :return: current user balance from exchange
        """
        raise NotImplementedError("get_balance is not implemented")

    async def get_symbol_prices(self,
                                symbol: str,
                                time_frame: common_enums.TimeFrames,
                                limit: int = None,
                                **kwargs: dict) -> typing.Optional[list]:
        """
        Return the candle history
        :param symbol: the symbol
        :param time_frame: the timeframe
        :param limit: the history limit size
        :return: the symbol candle history
        """
        raise NotImplementedError("get_symbol_prices is not implemented")

    async def get_kline_price(self,
                              symbol: str,
                              time_frame: common_enums.TimeFrames,
                              **kwargs: dict) -> typing.Optional[list]:
        """
        Return the symbol current kline dict
        :param symbol: the symbol
        :param time_frame: the timeframe
        :return: the symbol current klint
        """
        raise NotImplementedError("get_symbol_prices is not implemented")

    async def get_order_book(self, symbol: str, limit: int = 5, **kwargs: dict) -> typing.Optional[dict]:
        """
        Return the current symbol order book snapshot
        :param symbol: the symbol
        :param limit: the order book size
        :return: the order book snapshot
        """
        raise NotImplementedError("get_order_book is not implemented")

    async def get_recent_trades(self, symbol: str, limit: int = 50, **kwargs: dict) -> typing.Optional[list]:
        """
        Return the last "limit" recent trades for the specified symbol
        :param symbol: the symbol
        :param limit: the recent trade history size
        :return: the recent trade history for the symbol
        """
        raise NotImplementedError("get_recent_trades is not implemented")

    async def get_price_ticker(self, symbol: str, **kwargs: dict) -> typing.Optional[dict]:
        """
        Get the symbol ticker from the exchange
        :param symbol: the symbol
        :return: the symbol ticker
        """
        raise NotImplementedError("get_price_ticker is not implemented")

    async def get_all_currencies_price_ticker(self, **kwargs: dict) -> typing.Optional[list]:
        """
        Get all exchange currency tickers
        :return: the list of exchange currencies tickers
        """
        raise NotImplementedError("get_all_currencies_price_ticker is not implemented")

    async def get_order(self, exchange_order_id: str, symbol: str = None, **kwargs: dict) -> dict:
        """
        Get the order data from the exchange
        :param exchange_order_id: the order id on exchange
        :param symbol: the order symbol
        :return: the order data
        """
        raise NotImplementedError("get_order is not implemented")

    async def get_all_orders(self, symbol: str = None, since: int = None,
                             limit: int = None, **kwargs: dict) -> list:
        """
        Get the current user order list
        :param symbol: the order symbol
        :param since: the starting timestamp
        :param limit: the list limit size
        :return: the user order list
        """
        raise NotImplementedError("get_all_orders is not implemented")

    async def get_open_orders(self, symbol: str = None, since: int = None,
                              limit: int = None, **kwargs: dict) -> list:
        """
        Get the current user open order list
        :param symbol: the order symbol
        :param since: the starting timestamp
        :param limit: the list limit size
        :return: the user open order list
        """
        raise NotImplementedError("get_open_orders is not implemented")

    async def get_closed_orders(self, symbol: str = None, since: int = None,
                                limit: int = None, **kwargs: dict) -> list:
        """
        Get the user closed order list
        :param symbol: the order symbol
        :param since: the starting timestamp
        :param limit: the list limit size
        :return: the user closed order list
        """
        raise NotImplementedError("get_closed_orders is not implemented")

    async def get_my_recent_trades(self, symbol: str = None, since: int = None,
                                   limit: int = None, **kwargs: dict) -> list:
        """
        Get the user recent trades
        :param symbol: trades symbol
        :param since: the trade history starting timestamp
        :param limit: the history limit size
        :return: the user trades history list
        """
        raise NotImplementedError("get_my_recent_trades is not implemented")
        
    async def get_leverage_tiers(self, symbols: list = None, **kwargs: dict)-> dict:
        """
        :param symbols: the symbols or None
        :return: the current leverage tiers by symbols
        """
        raise NotImplementedError("get_leverage_tiers is not implemented")

    async def cancel_order(
            self, exchange_order_id: str, symbol: str, order_type: enums.TraderOrderType, **kwargs: dict
    ) -> enums.OrderStatus:
        """
        Cancel a order on the exchange
        :param exchange_order_id: the order id on exchange
        :param symbol: the order symbol
        :param order_type: the type of the order
        :return: True if the order is successfully cancelled
        """
        raise NotImplementedError("cancel_order is not implemented")

    async def create_order(self, order_type: enums.TraderOrderType, symbol: str, quantity: decimal.Decimal,
                           price: decimal.Decimal = None, stop_price: decimal.Decimal = None,
                           side: enums.TradeOrderSide = None, current_price: decimal.Decimal = None,
                           reduce_only: bool = False, params: dict = None) \
            -> typing.Optional[dict]:
        """
        Create a order on the exchange
        :param order_type: the order type
        :param symbol: the order symbol
        :param quantity: the order quantity
        :param price: the order price
        :param stop_price: the order stop price
        :param side: the order side
        :param current_price: the symbol current price
        :param params: the order request params
        :return: the created order dict
        """
        raise NotImplementedError("create_order is not implemented")

    def get_order_additional_params(self, order) -> dict:
        """
        Returns a dict with exchange specific additional parameters to set before sending the order
        :param order: the order instance wrapping orders details
        :return: the params dict
        """
        return {}

    def supports_bundled_order_on_order_creation(self, base_order, bundled_order_type) -> bool:
        """
        Returns True when this exchange supports orders created upon other orders fill (ex: a stop loss created at
        the same time as a buy order)
        :param base_order: the first order of the combo
        :param bundled_order_type: the type of the order that we try to bundle with base_order
        :return: True if an order of type tied_order_type can be pushed alongside base_order to the exchange
        in one request
        """
        try:
            return bundled_order_type in self.get_supported_elements(
                enums.ExchangeSupportedElements.SUPPORTED_BUNDLED_ORDERS
            )[base_order.order_type]
        except KeyError:
            return False

    def get_bundled_order_parameters(self, order, stop_loss_price=None, take_profit_price=None) -> dict:
        """
        Returns the updated params when this exchange supports orders created upon other orders fill
        (ex: a stop loss created at the same time as a buy order)
        :param order: the initial order
        :param stop_loss_price: the bundled order stop_loss price
        :param take_profit_price: the bundled order take_profit price
        :return: A dict with the necessary parameters to create the bundled order on exchange alongside the
        base order in one request
        """
        raise NotImplementedError("get_bundled_order_parameters is not implemented")

    def is_supported_order_type(self, order_type):
        """
        Check if the order type is supported by the current exchange instance
        Should be used to know if we should simulate this order or create it on the exchange
        :param order_type: the order type, should be a member of enums.TraderOrderType
        :return: True if the order type is supported by the exchange, else False
        """
        return order_type not in self.get_supported_elements(enums.ExchangeSupportedElements.UNSUPPORTED_ORDERS)

    def get_trade_fee(self, symbol, order_type, quantity, price, taker_or_maker):
        """
        Calculates fees resulting to a trade
        :param symbol: the symbol
        :param order_type: the order type
        :param quantity: the trade quantity
        :param price: the trade price
        :param taker_or_maker: if the trade was taker or maker
        :return: the trade fees
        """
        raise NotImplementedError("get_trade_fee is not implemented")

    def get_fees(self, symbol):
        """
        :param symbol: the symbol
        :return: the symbol fees dict
        """
        raise NotImplementedError("get_fees is not implemented")

    def get_pair_from_exchange(self, pair) -> str:
        """
        :param pair: the pair
        :return: the symbol associated to the pair
        """
        raise NotImplementedError("get_pair_from_exchange is not implemented")

    def get_split_pair_from_exchange(self, pair) -> (str, str):
        """
        :param pair: the pair
        :return: the currency, market tuple associated to the pair
        """
        raise NotImplementedError("get_split_pair_from_exchange is not implemented")

    def get_exchange_pair(self, pair) -> str:
        """
        :param pair: the pair
        :return: the exchange pair from an uniformized symbol
        """
        raise NotImplementedError("get_exchange_pair is not implemented")

    def get_pair_cryptocurrency(self, pair) -> str:
        """
        :param pair: the pair
        :return: the currency associated to the input pair
        """
        raise NotImplementedError("get_pair_cryptocurrency is not implemented")

    def get_default_balance(self):
        """
        :return: the default balance dict from exchange
        """
        raise NotImplementedError("get_default_balance is not implemented")

    def get_rate_limit(self):
        """
        :return: the exchange rate limit
        """
        raise NotImplementedError("get_default_balance is not implemented")

    async def switch_to_account(self, account_type: enums.AccountTypes):
        """
        Request to switch account from exchange
        :param account_type: the account destination
        """
        raise NotImplementedError("switch_to_account is not available on this exchange")

    async def get_sub_account_list(self):
        """
        :return: the exchange sub account list if supported by the exchange
        """
        raise NotImplementedError("get_sub_account_list is not available on this exchange")

    async def retry_till_success(self, timeout, request_func, *args, **kwargs):
        return await self._retry_until(timeout, 0, request_func, *args, **kwargs)

    async def retry_n_time(self, n_times, request_func, *args, **kwargs):
        return await self._retry_until(0, n_times, request_func, *args, **kwargs)

    async def _retry_until(self, timeout, n_times, request_func, *args, **kwargs):
        t0 = time.time()
        minimal_interval = 0.1
        attempt = 1
        while (timeout != 0 and time.time() - t0 < timeout) or (n_times != 0 and attempt <= n_times + 1):
            last_request_time = time.time()
            try:
                result = await request_func(*args, **kwargs)
                if attempt > 1:
                    self.logger.debug(f"Request retrier success for {request_func.__name__} after {attempt} attempts")
                return result
            except errors.FailedRequest as err:
                self.logger.debug(f"Request retrier failed for {request_func.__name__}({args} {kwargs}) (attempt {attempt}) ({err})")
                if time.time() - last_request_time < minimal_interval:
                    await asyncio.sleep(minimal_interval)
                attempt += 1
        raise errors.FailedRequest(f"Failed to successfully run {request_func.__name__} request after {attempt} "
                                   f"attempts.")

    """
    Parsers
    """

    def parse_balance(self, balance):
        """
        :param balance: the balance dict
        :return: the uniformized balance dict
        """
        raise NotImplementedError("parse_balance is not implemented")

    def parse_trade(self, trade):
        """
        :param trade: the trade dict
        :return: the uniformized trade dict
        """
        raise NotImplementedError("parse_trade is not implemented")

    def parse_order(self, order):
        """
        :param order: the order dict
        :return: the uniformized order dict
        """
        raise NotImplementedError("parse_order is not implemented")

    def parse_ticker(self, ticker):
        """
        :param ticker: the ticker dict
        :return: the uniformized ticker dict
        """
        raise NotImplementedError("parse_ticker is not implemented")

    def parse_ohlcv(self, ohlcv):
        """
        :param ohlcv: the ohlcv dict
        :return: the uniformized ohlcv dict
        """
        raise NotImplementedError("parse_ohlcv is not implemented")

    def parse_order_book(self, order_book):
        """
        :param order_book: the order book data
        :return: the uniformized order book data
        """
        raise NotImplementedError("parse_order_book is not implemented")

    def parse_order_book_ticker(self, order_book_ticker):
        """
        :param order_book_ticker: the order book ticker
        :return: the uniformized order book ticker
        """
        raise NotImplementedError("parse_order_book_ticker is not implemented")

    def parse_timestamp(self, data_dict, timestamp_key, default_value=None, ms=False):
        """
        Uniformize a raw timestamp from an input dict
        :param data_dict: the input dict
        :param timestamp_key: the timestamp dict in the input dict
        :param default_value: the default timestamp value
        :param ms: when True, return the timestamp in milliseconds
        :return: the uniformized timestamp
        """
        raise NotImplementedError("parse_timestamp is not implemented")

    def parse_currency(self, currency):
        """
        :param currency: the raw currency
        :return: the uniformized currency
        """
        raise NotImplementedError("parse_currency is not implemented")

    def parse_order_id(self, order):
        """
        :param order: the order dict
        :return: the order id
        """
        raise NotImplementedError("parse_order_id is not implemented")

    def parse_order_symbol(self, order):
        """
        :param order: the order dict
        :return: the order symbol
        """
        raise NotImplementedError("parse_order_symbol is not implemented")

    def parse_side(self, side):
        """
        :param side: the raw side
        :return: the TradeOrderSide related to the side
        """
        raise NotImplementedError("parse_side is not implemented")

    def parse_account(self, account):
        """
        :param account: the raw account
        :return: the AccountTypes related to the account
        """
        raise NotImplementedError("parse_account is not implemented")

    """
    Uniformization
    """

    def need_to_uniformize_timestamp(self, timestamp):
        """
        Return True if the timestamp should be uniformized
        :param timestamp: the timestamp to check
        :return: True if the timestamp should be uniformized
        """
        return not timestamp_util.is_valid_timestamp(timestamp)

    def get_uniformized_timestamp(self, timestamp):
        """
        Uniformize a timestamp
        :param timestamp: the timestamp to uniform
        :return: the timestamp uniformized
        """
        if self.need_to_uniformize_timestamp(timestamp):
            return self.get_uniform_timestamp(timestamp)
        return timestamp

    def uniformize_candles_if_necessary(self, candle_or_candles):
        """
        Uniform timestamps of a list of candles or a candle
        :param candle_or_candles: a list of candles or a candle to be uniformized
        :return: the list of candles or the candle uniformized
        """
        if candle_or_candles:  # TODO improve
            if isinstance(candle_or_candles[0], list):
                if self.need_to_uniformize_timestamp(
                        candle_or_candles[0][common_enums.PriceIndexes.IND_PRICE_TIME.value]):
                    self.uniformize_candles_timestamps(candle_or_candles)
            else:
                if self.need_to_uniformize_timestamp(candle_or_candles[common_enums.PriceIndexes.IND_PRICE_TIME.value]):
                    self.uniformize_candle_timestamps(candle_or_candles)
        return candle_or_candles

    def uniformize_candles_timestamps(self, candles):
        """
        Uniformize a list candle timestamps
        :param candles: the list of candles to uniformize
        """
        for candle in candles:
            self.uniformize_candle_timestamps(candle)

    def uniformize_candle_timestamps(self, candle):
        """
        Uniformize a candle timestamp
        :param candle: the candle to uniformize
        """
        candle[common_enums.PriceIndexes.IND_PRICE_TIME.value] = \
            self.get_uniform_timestamp(candle[common_enums.PriceIndexes.IND_PRICE_TIME.value])

    def get_candle_since_timestamp(self, time_frame, count):
        """
        :param time_frame: the time frame to use
        :param count: the number of candle
        :return: the timestamp since "count" candles
        """
        return self.get_exchange_current_time() - (common_enums.TimeFramesMinutes[time_frame]
                                                   * octobot_commons.constants.MSECONDS_TO_MINUTE
                                                   * count)

    def get_max_handled_pair_with_time_frame(self) -> int:
        """
        :return: the maximum number of simultaneous pairs * time_frame that this exchange can handle.
        """
        return self.connector.get_max_handled_pair_with_time_frame()

    def get_additional_connector_config(self):
        """
        :return: a dict with additional elements to give to the connector constructor
        """
        return {}

    @classmethod
    def is_configurable(cls):
        """
        Override if the exchange is allowed to be configured
        """
        return False

    def should_log_on_ddos_exception(self, _) -> bool:
        """
        Override when necessary
        """
        return True

    def log_order_creation_error(self, error, order_type, symbol, quantity, price, stop_price):
        order_desc = f"order_type: {order_type}, symbol: {symbol}, quantity: {str(quantity)}, price: {str(price)}," \
                     f" stop_price: {str(stop_price)}"
        self.logger.error(f"Failed to create order : {error.__class__.__name__} {error}: ({order_desc})")

    def handle_token_error(self, error):
        self.logger.error(f"Exchange configuration is invalid : please check your configuration ! "
                          f"({error.__class__.__name__}: {error})")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.databases as databases
import octobot_commons.constants as common_constants
import octobot_tentacles_manager.api as tentacles_manager_api
import octobot_trading.constants as constants
import octobot_trading.storage as storage
import octobot_trading.modes.script_keywords.basic_keywords as basic_keywords


def get_required_candles_count(trading_mode_class, tentacles_setup_config):
    return tentacles_manager_api.get_tentacle_config(tentacles_setup_config, trading_mode_class).get(
        constants.CONFIG_CANDLES_HISTORY_SIZE_KEY,
        common_constants.DEFAULT_IGNORED_VALUE
    )


async def clear_simulated_orders_cache(trading_mode):
    await basic_keywords.clear_orders_cache(
        databases.RunDatabasesProvider.instance().get_orders_db(
            trading_mode.bot_id,
            storage.get_account_type_suffix_from_exchange_manager(trading_mode.exchange_manager),
            trading_mode.exchange_manager.exchange_name
        )
    )


async def clear_plotting_cache(trading_mode):
    await basic_keywords.clear_symbol_plot_cache(
        databases.RunDatabasesProvider.instance().get_symbol_db(
            trading_mode.bot_id,
            trading_mode.exchange_manager.exchange_name, trading_mode.symbol
        )
    )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_trading.personal_data.orders.order as order_class


class LimitOrder(order_class.Order):
    def __init__(self, trader, side=enums.TradeOrderSide.BUY):
        super().__init__(trader, side)
        self.limit_price_hit_event = None
        self.wait_for_hit_event_task = None
        self.trigger_above = self.side is enums.TradeOrderSide.SELL
        self.allow_instant_fill = True

    def update_from_raw(self, raw_order):
        updated = super(LimitOrder, self).update_from_raw(raw_order)
        if enums.ExchangeConstantsOrderColumns.TRIGGER_ABOVE.value in raw_order \
                and self.trigger_above != raw_order[enums.ExchangeConstantsOrderColumns.TRIGGER_ABOVE.value]:
            updated = True
            self.trigger_above = raw_order[enums.ExchangeConstantsOrderColumns.TRIGGER_ABOVE.value]
        return updated

    async def update_order_status(self, force_refresh=False):
        if self.limit_price_hit_event is None:
            self._create_hit_event(self.creation_time)

        if self.wait_for_hit_event_task is None and self.limit_price_hit_event is not None:
            if self.limit_price_hit_event.is_set():
                # order should be filled instantly
                await self.on_fill()
            else:
                # order will be filled when conditions are met
                self._create_hit_task()

    def _on_origin_price_change(self, previous_price, price_time):
        if previous_price is not constants.ZERO:
            # no need to reset events if previous price was 0 (unset)
            self._reset_events(price_time)

    def _create_hit_event(self, price_time):
        self.limit_price_hit_event = self.exchange_manager.exchange_symbols_data.\
            get_exchange_symbol_data(self.symbol).price_events_manager.\
            new_event(self.origin_price, price_time, self.trigger_above, self.allow_instant_fill)

    def _create_hit_task(self):
        self.wait_for_hit_event_task = asyncio.create_task(self.wait_for_price_hit())

    def _reset_events(self, price_time):
        """
        Reset events and tasks
        """
        self._clear_event_and_tasks()
        self._create_hit_event(price_time)
        self._create_hit_task()

    def _clear_event_and_tasks(self):
        if self.wait_for_hit_event_task is not None:
            if not self.limit_price_hit_event.is_set():
                self.wait_for_hit_event_task.cancel()
            self.wait_for_hit_event_task = None
        if self.limit_price_hit_event is not None:
            self.exchange_manager.exchange_symbols_data. \
                get_exchange_symbol_data(self.symbol).price_events_manager.remove_event(self.limit_price_hit_event)

    async def wait_for_price_hit(self):
        await asyncio.wait_for(self.limit_price_hit_event.wait(), timeout=None)
        await self.on_fill()

    def _filled_maker_or_taker(self):
        return enums.ExchangeConstantsMarketPropertyColumns.MAKER.value

    def on_fill_actions(self):
        self.taker_or_maker = self._filled_maker_or_taker()
        self.filled_price = self.origin_price
        self.filled_quantity = self.origin_quantity
        self._update_total_cost()
        order_class.Order.on_fill_actions(self)

    def clear(self):
        self._clear_event_and_tasks()
        order_class.Order.clear(self)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.constants as constants
import octobot_trading.errors as errors
import octobot_trading.util as util


def get_profitability_stats(exchange_manager) -> tuple:
    port_profit = exchange_manager.exchange_personal_data.portfolio_manager.portfolio_profitability
    return port_profit.profitability, \
        port_profit.profitability_percent, \
        port_profit.profitability_diff, \
        port_profit.market_profitability_percent, \
        port_profit.initial_portfolio_current_profitability


def get_origin_portfolio_value(exchange_manager) -> float:
    return exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.portfolio_origin_value


def get_current_portfolio_value(exchange_manager) -> float:
    return exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.portfolio_current_value


def get_current_crypto_currency_value(exchange_manager, currency) -> decimal.Decimal:
    try:
        return exchange_manager.exchange_personal_data.portfolio_manager. \
            portfolio_value_holder.current_crypto_currencies_values[currency]
    except KeyError:
        try:
            # try using last prices by trading pair
            return exchange_manager.exchange_personal_data.portfolio_manager. \
                portfolio_value_holder.value_converter.convert_currency_value_using_last_prices(
                    constants.ONE,
                    currency,
                    exchange_manager.exchange_personal_data.portfolio_manager.reference_market
                )
        except errors.MissingPriceDataError as err:
            raise KeyError from err


def get_currency_ref_market_value(exchange_manager, currency):
    try:
        return exchange_manager.exchange_personal_data.portfolio_manager. \
            portfolio_value_holder.value_converter.evaluate_value(currency, constants.ONE) or None
    except errors.MissingPriceDataError:
        return None


def get_current_holdings_values(exchange_manager) -> dict:
    return exchange_manager.exchange_personal_data.portfolio_manager.\
        portfolio_value_holder.get_current_holdings_values()


def get_reference_market(config) -> str:
    return util.get_reference_market(config)


def get_initializing_currencies_prices(exchange_manager) -> set:
    return set() if exchange_manager.exchange_personal_data.portfolio_manager is None else \
        exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        value_converter.initializing_symbol_prices



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import ccxt.async_support
import octobot_trading.exchanges.util as util
import octobot_trading.exchanges.connectors.ccxt.ccxt_client_util as ccxt_client_util


def parse_markets(exchange_name, forced_markets: list[util.SymbolDetails]) -> dict:
    client = getattr(ccxt.async_support, exchange_name)()
    ccxt_client_util.set_markets_from_forced_markets(client, forced_markets)
    return client.markets


def get_fees(market_status) -> dict:
    return ccxt_client_util.get_fees(market_status)
